#  ========================================================================
#  DEVise Data Visualization Software
#  (c) Copyright 1992-2001
#  By the DEVise Development Group
#  Madison, Wisconsin
#  All Rights Reserved.
#  ========================================================================
#
#  Under no circumstances is this software to be copied, distributed,
#  or altered in any way without prior permission from the DEVise
#  Development Group.

############################################################

#  $Id: window.tcl,v 1.37 2001/03/15 20:34:27 wenger Exp $

#  $Log: window.tcl,v $
#  Revision 1.37  2001/03/15 20:34:27  wenger
#  Reading composite file /afs/cs.wisc.edu/p/devise/ext5/wenger/devise.dev2/solarisFixed bug 645 (problem with window duplication).
#
#  Revision 1.36  1999/07/14 18:43:05  wenger
#  Added the capability to have axes without ticks and tick labels.
#
#  Revision 1.35  1999/02/11 19:55:06  wenger
#  Merged newpile_br through newpile_br_1 (new PileStack class controls
#  pile and stacks, allows non-linked piles; various other improvements
#  to pile-related code).
#
#  Revision 1.34.2.2  1999/02/11 18:24:31  wenger
#  PileStack objects are now fully working (allowing non-linked piles) except
#  for a couple of minor bugs; new PileStack state is saved to session files;
#  piles and stacks in old session files are dealt with reasonably well;
#  incremented version number; added some debug code.
#
#  Revision 1.34.2.1  1998/12/29 17:25:23  wenger
#  First version of new PileStack objects implemented -- allows piles without
#  pile links.  Can't be saved or restored in session files yet.
#
#  Revision 1.34  1998/10/20 19:46:36  wenger
#  Mapping dialog now displays the view's TData name; "Next in Pile" button
#  in mapping dialog allows user to edit the mappings of all views in a pile
#  without actually flipping them; user has the option to show all view names;
#  new GUI to display info about all links and cursors; added API and GUI for
#  count mappings.
#
#  Revision 1.33  1998/03/05 20:36:30  wenger
#  Fixed bugs 304 and 309 (problems with view colors); fixed a few other
#  problems related to *ClassInfo classes.
#
#  Revision 1.32  1997/07/15 14:59:30  wenger
#  Fixed a bug in the view destroying Tcl code; user can now destroy a window
#  that contains views (DEVise asks for confirmation, destroys the views).
#
#  Revision 1.31  1997/06/18 15:33:36  wenger
#  Fixed bug 177; improved workaround of bug 137; incremented version
#  number (because of Unidata being added).
#
#  Revision 1.30  1997/05/28 15:39:56  wenger
#  Merged Shilpa's layout manager code through the layout_mgr_branch_2 tag.
#
#  Revision 1.29  1997/05/08 00:14:38  wenger
#  Removed layout manager GUI.
#
#  Revision 1.28.4.1  1997/05/22 03:07:47  ssl
#  DEVise views can be created from LM.
#
#  Revision 1.28  1997/02/14 16:48:10  wenger
#  Merged 1.3 branch thru rel_1_3_1 tag back into the main CVS trunk.
#
#  Revision 1.27  1997/02/03 20:02:03  ssl
#  Added interface for negative record links and user defined layout mode
#
#  Revision 1.26.4.5  1997/02/11 16:00:24  wenger
#  Commented out some debug code in the links stuff; added standard header
#  to dialogInfo.tk.
#
#  Revision 1.26.4.4  1997/02/11 01:17:40  ssl
#  Cleaned up the UI for piled views
#  1) Made pile links invisible to user
#  2) Added create/destroy link options to the link menu
#  3) Enhanced the link info window to show all info about a link (type, views,
#     master, link params )
#  4) Pile links get removed when the pile is unpiled
#  5) Set/Reset Master now only shows list of record links
#
#  Revision 1.26.4.3  1997/02/09 17:04:33  wenger
#  Fixed bug 097 (duplicating a window now creates a new set of mappings for
#  the views in the new window).
#
#  Revision 1.26.4.2  1997/02/08 02:08:43  ssl
#  1)Fixed mapping dialog to update when selected view changes
#  2)Removed  OK buttons and renamed Cancel to Close
#  3)Added a Flip button which allows user to flip views whenever the view
#    is a pile. NOT ENABLED FOR STACKs or other views.
#
#  Revision 1.26.4.1  1997/02/07 15:21:48  wenger
#  Updated Devise version to 1.3.1; fixed bug 148 (GUI now forces unique
#  window names); added axis toggling and color selections to Window menu;
#  other minor fixes to GUI; show command to Tasvir now requests image to
#  be shown all at once.
#
#  Revision 1.26  1997/01/27 22:39:38  wenger
#  Fixed workaround to bug 137 so it doesn't mess things up in batch mode.
#
#  Revision 1.25  1997/01/27 20:15:25  wenger
#  Workaround to bug 137: disables Stack Control dialog buttons while drawing.
#
#  Revision 1.24  1997/01/23 21:47:45  wenger
#  Fixed a couple more OK/Cancel button pairs.
#
#  Revision 1.23  1997/01/23 19:28:27  jussi
#  Removed debugging messages. Fixed bug #126.
#
#  Revision 1.22  1997/01/23 16:02:24  jussi
#  Removed debugging message.
#
#  Revision 1.21  1997/01/23 16:02:12  jussi
#  Added code that raises the top view in a stack when stack
#  order is rotated.
#
#  Revision 1.20  1997/01/23 00:08:12  jussi
#  Removed references to stackWinOpened. Removed Rotate button.
#
#  Revision 1.19  1997/01/22 20:14:00  wenger
#  Removed other non-functional user interface components (includes workaround
#  for bug 127); fixed a number of OK/Cancel button positions.
#
#  Revision 1.18  1997/01/22 14:40:34  wenger
#  Added devise.etk to release scripts; removed merged view buttons from
#  Stack Control dialog.
#
#  Revision 1.17  1997/01/17 20:32:32  wenger
#  Fixed bugs 088, 121, 122; put workaround in place for bug 123; added
#  simulation of XOR drawing in PSWindowRep; removed diagnostic output
#  from Tcl/Tk code; removed (at least for now) the ETk interface from
#  the cslib versions of WindowRep classes so that the cslib will link
#  okay; cslib server now tests XOR drawing.
#
#  Revision 1.16  1997/01/08 18:56:51  wenger
#  Added 'View List' selection to window menu (gives a list of the views
#  in the window containing the selected view, especially useful for
#  piled views); fixed OK/Cancel in Remove Window dialog.
#
#  Revision 1.15  1996/11/26 17:25:22  ssl
#  added support for stacked views
#
#  Revision 1.14  1996/08/03 15:51:52  jussi
#  3D view parameters are preserved when a merged view is split.
#
#  Revision 1.13  1996/08/03 15:34:39  jussi
#  When views are merged, the 3D parameters of the currently selected
#  view will determine the 3D parameters of the composite view.
#
#  Revision 1.12  1996/07/23 17:27:18  jussi
#  Added support for piled views.
#
#  Revision 1.11  1996/07/21 14:59:45  jussi
#  Code no longer inserts an "X" when rotating legend-less mappings
#  in a merged view.
#
#  Revision 1.10  1996/07/08 17:08:38  jussi
#  The mapping legend (label) is now copied in DupWindow.
#
#  Revision 1.9  1996/07/01 19:32:28  jussi
#  The 3D camera location is now duplicated in Window/Duplicate.
#
#  Revision 1.8  1996/05/31 19:14:30  jussi
#  Fixed calls to DEVise removeView; there is no return value from
#  the call.
#
#  Revision 1.7  1996/05/11 03:01:13  jussi
#  Minor fix in DoActualCreateWindow.
#
#  Revision 1.6  1996/05/09 18:15:21  kmurli
#  No changes, just inserted some debugging print statements and removed them
#
#  Revision 1.5  1996/05/07 16:48:12  jussi
#  Rotation of mappings not done if there is just one mappings
#  in a view.
#
#  Revision 1.4  1996/04/23 20:36:46  jussi
#  Replaced Unknown view legend with X.
#
#  Revision 1.3  1996/04/15 15:05:39  jussi
#  Rewrote WindowMerge and WindowSplit code.
#
#  Revision 1.2  1996/04/14 00:07:08  jussi
#  DupWindow now copies all links of the source view into the new view.
#
#  Revision 1.1  1996/04/11 18:13:07  jussi
#  Initial revision.
#

############################################################

set stackEnabled 1

proc RemoveWindow {win} {
    global dialogListVar

    if {$win == ""} {
	set winSet [ WinSet ]
	set answer [ dialogList .window "Select Window" \
		"Select window to destroy" "" "" \
		{ OK Cancel } $winSet ]
	if {$answer == 1  || $dialogListVar(selected) == ""} {
	    return 
	}
	set win $dialogListVar(selected)
    }

    set views [DEVise getWinViews $win]
    if {$views != ""} {
	set answer [dialog .viewsInWindow "Views Still In Window" \
		"Window contains views.  Destroy window and all views, or cancel." \
		"" 1 Destroy Cancel]
	if {$answer == 1} {
	    return
	}
	foreach view $views {
	    DestroyView $view
	}
    }
    
    DEVise destroy $win
}

############################################################

proc SetWindowLayout {win} {
    global curView layoutWin layoutOption verRequested horRequested

    if {$win == ""} {
	if {![CurrentView]} {
	    return
	}
	set win [DEVise getViewWin $curView]
    }

    if {[WindowVisible .setLayout]} {
	return
    }

    set layoutWin $win

    toplevel .setLayout
    wm title .setLayout "Set Window Layout"
    wm geometry .setLayout +50+50

    frame .setLayout.top
    frame .setLayout.bot
    pack .setLayout.top -side top -pady 3m -fill both -expand 1
    pack .setLayout.bot -side top -pady 5m -fill x

    frame .setLayout.row1
    frame .setLayout.row2
    frame .setLayout.row3
    frame .setLayout.row4
    frame .setLayout.row5

    pack .setLayout.row1 .setLayout.row2 .setLayout.row3 .setLayout.row4 \
	    .setLayout.row5 -in .setLayout.top -side top -fill x

    frame .setLayout.bot.but
    pack .setLayout.bot.but -side top

    radiobutton .setLayout.row1.b -text Automatic -variable layoutOption \
	    -value auto -width 12 -anchor w -command {
	.setLayout.row2.e configure -state disabled
	.setLayout.row3.e configure -state disabled
#	.setLayout.row5.edit configure -state disabled
    }
    pack .setLayout.row1.b -side left

    radiobutton .setLayout.row2.b -text Vertical \
	    -variable layoutOption -value fixver -width 12 -anchor w \
	    -command {
	.setLayout.row2.e configure -state normal
	.setLayout.row3.e configure -state disabled
#	.setLayout.row5.edit configure -state disabled
    }
    label .setLayout.row2.l -text "Width:" -width 8
    entry .setLayout.row2.e -relief sunken -textvariable horRequested \
	    -width 5
    pack .setLayout.row2.b .setLayout.row2.l .setLayout.row2.e -side left

    radiobutton .setLayout.row3.b -text Horizontal \
	    -variable layoutOption -value fixhor -width 12 -anchor w \
	    -command {
	.setLayout.row2.e configure -state disabled
	.setLayout.row3.e configure -state normal
#	.setLayout.row5.edit configure -state disabled
    }
    label .setLayout.row3.l -text "Height:" -width 8
    entry .setLayout.row3.e -relief sunken -textvariable verRequested \
	    -width 5
    pack .setLayout.row3.b .setLayout.row3.l .setLayout.row3.e -side left

# Disabled Stacked here because windows should now be stacked only via
# the PileStack class.  RKW 1999-02-11.
#    radiobutton .setLayout.row4.b -text Stacked -variable layoutOption \
#	    -value stacked -width 12 -anchor w -command {
#	.setLayout.row2.e configure -state disabled
#	.setLayout.row3.e configure -state disabled
##	.setLayout.row5.edit configure -state disabled
#    }
#    pack .setLayout.row4.b -side left
#    radiobutton .setLayout.row5.b -text UserDefined  -variable layoutOption \
#	    -value  custom  -width 12 -anchor w -command {
#	.setLayout.row2.e configure -state disabled
#	.setLayout.row3.e configure -state disabled
#	.setLayout.row5.edit configure -state normal
#    }
#    button .setLayout.row5.edit -text "Editor" -command { RootLayout } 
#    pack .setLayout.row5.b .setLayout.row5.edit -side left

    set verRequested 1
    set horRequested 1
    set layoutOption auto
    .setLayout.row2.e configure -state disabled
    .setLayout.row3.e configure -state disabled
#    .setLayout.row5.edit configure -state disabled

    set layout [DEVise getWindowLayout $layoutWin]
    if {[lindex $layout 0] >= 1} {
	set verRequested [lindex $layout 0]
    } elseif {[lindex $layout 1] >= 1} {
	set horRequested [lindex $layout 1]
    }
    if {[lindex $layout 2]} {
	set layoutOption stacked
    } elseif {[lindex $layout 0] >= 1} {
	set layoutOption fixhor
	.setLayout.row3.e configure -state normal
    } elseif {[lindex $layout 1] >= 1} {
	set layoutOption fixver
	.setLayout.row2.e configure -state normal
    } elseif {[lindex $layout 0] == 0 && [lindex $layout 1] == 0 } {
	set layoutOption custom
#	.setLayout.row5.edit configure -state normal
    }

    button .setLayout.bot.but.ok -text OK -width 10 -command {
	if {$layoutOption == "auto"} {
	    DEVise setWindowLayout $layoutWin -1 -1 0
	} elseif {$layoutOption == "fixver"} {
	    if {$horRequested < 1} { set horRequested 1 }
	    DEVise setWindowLayout $layoutWin -1 $horRequested 0
	} elseif {$layoutOption == "fixhor"} {
	    if {$verRequested < 1} { set verRequested 1 }
	    DEVise setWindowLayout $layoutWin $verRequested -1 0
        } elseif {$layoutOption == "custom"} {
	    DEVise setWindowLayout $layoutWin 0  0  0
	} else {
# Disabled Stacked here because windows should now be stacked only via
# the PileStack class.  RKW 1999-02-11.
#	    DEVise setWindowLayout $layoutWin -1 -1 1
	}
	destroy .setLayout
    }
    button .setLayout.bot.but.cancel -text Cancel -width 10 -command {
	destroy .setLayout
    }
    pack .setLayout.bot.but.ok .setLayout.bot.but.cancel -side left -padx 3m
}

############################################################

proc DupWindow {} {
    global curView

    if {![CurrentView]} {
	return
    }

    set win [DEVise getViewWin $curView]

    # Create new window
    set class [GetClass window $win]
    set params [DEVise getCreateParam window $class $win]
    set newWin [UniqueName $win]
    set newParam [linsert [lrange $params 1 end] 0 $newWin]
    eval DEVise create window $class $newParam

    set layout [DEVise getWindowLayout $win]
    eval DEVise setWindowLayout {$newWin} $layout

    set views [DEVise getWinViews $win]
    
    # Create mappings for views
    foreach view $views {
        DupView $view $newWin
    }

    DEVise setPileStackState $newWin [DEVise getPileStackState $win]
}

############################################################

# Duplicate a view.  view is the name of the view; window is the
# name of the window to put it in.
proc DupView {view window} {

    set gdata [lindex [DEVise getViewMappings $view] 0]
    set gdataClass [GetClass mapping $gdata]
    set gdataParam [DEVise getCreateParam mapping $gdataClass $gdata]
    set tdata [lindex [DEVise getCreateParam mapping $gdataClass $gdata] 0]

    set gdataParam [DEVise getCreateParam mapping $gdataClass $gdata]
    set newGdata [UniqueName "$tdata#$gdataClass"]

    DoActualViewCopy $view $tdata $gdata $newGdata $window
}

############################################################

proc DoSavePixmap {} {
    global curView

    if {![CurrentView]} {
	return
    }
    
    set file [DEVise open pixmap.tmp wb]
    DEVise savePixmap $curView $file
    DEVise close $file
}

############################################################

proc DoLoadPixmap {} {
    global curView

    if {![CurrentView]} {
	return
    }
    
    set file [DEVise open pixmap.tmp rb]
    DEVise loadPixmap $curView $file
    DEVise close $file
}

############################################################

# Create a new window
proc DoCreateWindow { message { createdFromLM 0 } args } {

    global dialogListVar

    set winTypes [ DEVise get window ]
#    puts " createdFromLM = $createdFromLM  types = $winTypes"

    if { [llength $winTypes] == 1 } {
	return [ DoActualCreateWindow [lindex $winTypes 0] \
		$createdFromLM $args ]
    }

    set answer [ dialogList .createWin "Create Window"  \
	    $message "" "" { OK Cancel } $winTypes ]
    if {$answer == 1 || $dialogListVar(selected) == ""} {
	return ""
    } else {
	return [ DoActualCreateWindow $dialogListVar(selected) ]
    }
}

############################################################

proc DoActualCreateWindow { winType {createdFromLM 1} args } {
    global dialogParamVar windowName

    set windowName [UniqueName DEViseWn0]
#    puts "args = $args"
    if { $createdFromLM } {    
	set paramNames [list \
		[list Name $windowName] \
		[list X [lindex [lindex $args 0] 0] ] \
		[list Y [lindex [lindex $args 0] 1] ] \
		[list Width [lindex [lindex $args 0] 2] ] \
		[list Height [lindex [lindex $args 0] 3] ] ]\
    } else {
	set paramNames [ DEVise getparam window $winType ]
	set nameParm [lindex $paramNames 0]
	set nameParm [lreplace $nameParm 1 1 $windowName]
	set paramNames [lreplace $paramNames 0 0 $nameParm]
    }
    
#    puts "$paramNames"
    set button [ dialogParam .windowParam "Create Window" \
	    "Enter window parameters"\
	    "" 0 { OK Cancel } $paramNames ]

    if { $button == 1} {
	return
    }

    set windowName [lindex $dialogParamVar(params) 0]
    set windowName [UniqueName $windowName]
    set dialogParamVar(params) [lreplace $dialogParamVar(params) 0 0 $windowName]

    set cmd "DEVise create window $winType $dialogParamVar(params)"
    set result [eval $cmd]
    if {$result == ""} {
	set but [ dialog .winError "Window Error" \
		"Can't create window using $winType" "" 0 OK ]
	return
    }
    return $result
}

############################################################

# Get a window 
proc DoGetWindow {} {
    global dialogListVar

    set winSet [ WinSet ]

    if { [ llength $winSet ] == 0 } {
	return [ DoCreateWindow "Select window type" ]
    }

    set answer [ dialogList .window "Select Window" \
	    "Select a window to display view in" "" "" \
	    { OK New Cancel } $winSet ]
    if { $answer == 1 } {
	return [ DoCreateWindow "Select window type" ]
    }

    if {$answer == 2  || $dialogListVar(selected) == ""} {
	return 
    }

    return $dialogListVar(selected)
}

############################################################

proc DoWindowStackControl {} {
    global curView

    if {[WindowVisible .stack]} {
	return
    }

    toplevel    .stack
    wm title    .stack "Stack Control"
    wm geometry .stack +100+100

    # build view name widget

    frame .stack.title
    pack .stack.title -side top -pady 1m -expand 1 -fill x

    label .stack.title.text -text "No View Selected"
    pack .stack.title.text -side top -pady 1m

    # build button bar widget

    frame .stack.bot
    frame .stack.bot.row1
    frame .stack.bot.row2
    frame .stack.bot.row1.but
    frame .stack.bot.row2.but

    button .stack.bot.row1.but.pile -text "Pile (linked)" -width 12 \
	    -command { DoWindowPile 1 }
    button .stack.bot.row2.but.unpile -text "Pile (unlinked)" -width 12 \
	    -command { DoWindowPile 0 }
    button .stack.bot.row1.but.stack -text "Stack" -width 12 \
	    -command { DoWindowStack }
    button .stack.bot.row2.but.unstack -text "Flatten" -width 12 \
	    -command { DoWindowFlatten }
    button .stack.bot.row1.but.flip -text "Flip" -width 12 \
	    -command { FlipStackedView }
    button .stack.bot.row2.but.close -text Close -width 12 \
	    -command { destroy .stack }

    pack .stack.bot.row1.but.pile \
            .stack.bot.row1.but.stack \
            .stack.bot.row1.but.flip \
	    -side left -padx 3m
#    pack .stack.bot.row1.but.pile \
#            .stack.bot.row1.but.stack \
#            .stack.bot.row1.but.edit \
#	    -side left -expand 1 -fill x -padx 3m

    pack .stack.bot.row2.but.unpile \
	    .stack.bot.row2.but.unstack \
            .stack.bot.row2.but.close \
	    -side left -expand 1 -fill x -padx 3m

    pack .stack.bot.row1.but -side top
    pack .stack.bot.row2.but -side top
    pack .stack.bot.row1 .stack.bot.row2 -side top -pady 1m
    pack .stack.bot -side top -pady 5m

    if {$curView != ""} {
	if { [DEVise getViewPileMode $curView] } { 
	    set win [DEVise getViewWin $curView]
	    set views [DEVise getWinViews $win]
	    set topView [lindex $views end]
	    .stack.title.text configure -text "Top View : $topView"
	} else {
	    .stack.title.text configure -text "Current View: $curView"
	}
    }
}

############################################################

proc EnableStackControl {} {
    global stackEnabled
    set stackEnabled 1

    if {![WindowExists .stack]} {
	return
    }

    .stack.bot.row1.but.pile config -state normal
    .stack.bot.row2.but.unpile config -state normal
    .stack.bot.row1.but.stack config -state normal
    .stack.bot.row2.but.unstack config -state normal
    .stack.bot.row1.but.flip config -state normal
}

############################################################

proc DisableStackControl {} {
    global stackEnabled
    set stackEnabled 0

    if {![WindowExists .stack]} {
	return
    }

    .stack.bot.row1.but.pile config -state disabled
    .stack.bot.row2.but.unpile config -state disabled
    .stack.bot.row1.but.stack config -state disabled
    .stack.bot.row2.but.unstack config -state disabled
    .stack.bot.row1.but.flip config -state disabled
}

############################################################

proc FlipStackedView {} {
    global curView stackEnabled

    if {![CurrentView]} {
	return
    }

    if {!$stackEnabled} {
	puts "Stack control currently disabled"
	return
    }

    set win [DEVise getViewWin $curView]

    DEVise flipPileStack $win
}

############################################################

proc DoWindowPile { link } {
    global curView stackEnabled

    if {![CurrentView]} {
	return
    }

    if {!$stackEnabled} {
	puts "Stack control currently disabled"
	return
    }

    # See PileStack.h for pileState values.
    if {$link} {
    	set pileState 4
    } else {
    	set pileState 3
    }
    DEVise setPileStackState [DEVise getViewWin $curView] $pileState
}

############################################################

proc DoWindowStack {} {
    global curView stackEnabled

    if {![CurrentView]} {
	return
    }

    if {!$stackEnabled} {
	puts "Stack control currently disabled"
	return
    }

    DEVise setPileStackState [DEVise getViewWin $curView] 2
}

############################################################

proc DoWindowFlatten {} {
    global curView stackEnabled

    if {![CurrentView]} {
	return
    }

    if {!$stackEnabled} {
	puts "Stack control currently disabled"
	return
    }

    DEVise setPileStackState [DEVise getViewWin $curView] 1
}

############################################################

# Show a list of the views contained in the window containing the currently-
# selected view.

proc WinViewList {} {
    global curView

    if {![CurrentView]} {
	return
    }

    set win [DEVise getViewWin $curView]

    set views [DEVise getWinViews $win]
    if {$views != ""} {
	dialogList .winViewList "View List" "Window $win Views" "" 0 {OK} $views
    }
}

############################################################

proc DoToggleAxisWindow { axis } {
    global curView

    if {![CurrentView]} {
	return
    }

    set stat [DEVise getAxisDisplay $curView $axis]
    set stat [expr !$stat]

    set viewList [DEVise getWinViews [DEVise getViewWin $curView]]

    foreach viewName $viewList {
        DEVise setAxisDisplay $viewName $axis $stat
    }
}

############################################################

proc DoToggleAxisTicksWindow { axis } {
    global curView

    if {![CurrentView]} {
	return
    }

    set enabled [DEVise getAxisTicks $curView $axis]
    set enabled [expr !$enabled]

    set viewList [DEVise getWinViews [DEVise getViewWin $curView]]

    foreach viewName $viewList {
        DEVise setAxisTicks $viewName $axis $enabled
    }
}

############################################################
